;;; mixmaster.el -- Emacs Mixmaster interface

;; Copyright (C) 2001 Erik Arneson

;; Author:        Erik Arneson <erik@aarg.net>
;; Maintainer:    erik@aarg.net
;; Created:       Sep-20-2001
;; Version:       0.5
;; Last-Modified: <2003-01-24 07:40:32 (erik)>
;; Keywords:      mail, encryption, mixmaster, remailer, anonymous

;; This file may be part of XEmacs.

;; XEmacs is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; XEmacs is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with XEmacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.
;;
;; $Id: mixmaster.el,v 1.3 2003/01/25 19:27:07 erik Exp $

;;{{{ Introduction

;; This is a set of utility functions for interfacing with Mixmaster
;; 3.0.  It has been tested with version 2.9beta31 and seems to work
;; alright.  I also can't make any guarantees that it will work with GNU
;; Emacs, as I am an XEmacs user and really never open that *other*
;; Emacs.  But there's nothing too weird in here, so it should be
;; alright.

;; This is very simple to use.  In fact, the hardest part is setting up
;; Mixmaster.  Once this is done, set `mixmaster-directory' to the
;; Mixmaster installation location, and something like the following to
;; your .emacs or init.el.

;; (add-hook 'mail-mode-hook 'mixmaster-minor-mode)

;; How should you use this with other mailers?  Well, you need to just
;; make sure that you use `mixmaster-send-it' as the
;; `send-mail-function' and you'll be all set.  Here's an example for
;; VM.

;; (define-key vm-mail-mode-map [(control c) m m]
;;   (lambda (&optional arg)
;;     (interactive "P")
;;     (let ((send-mail-function 'mixmaster-send-it))
;;       (vm-mail-send-and-exit arg))))

;; Of course, you can aways do it the easy way, like this:

;; (add-hook 'vm-mail-mode-hook 'mixmaster-vm-setup)

;; To configure GNUS to post via Mixmaster upon request, do this:

;; (add-hook 'message-mode-hook 'mixmaster-gnus-setup)

;;{{{ TODO:

;; 1. Fix GNUS support.  I think it's still broken.
;; 2. Check message sizes agains remailer capabilities.
;; 3. Clean the output buffer automatically after a specific period of time?

;;}}}

;;}}}
;;{{{ Requirements

(require 'mail-utils)
(require 'sendmail)

;; And what can I say?  I'm a Common Lisper at heart.
(require 'cl)

;;}}}
;;{{{ Customization

(defgroup mixmaster nil
  "Mixmaster mail and news interface"
  :group 'comm)

(defcustom mixmaster-directory (expand-file-name "~/Mix")
  "Mixmaster installation directory."
  :group 'mixmaster)

(defcustom mixmaster-binary (concat mixmaster-directory "/mix")
  "Path to the Mixmaster binary."
  :group 'mixmaster)

(defcustom mixmaster-min-reliable nil
  "Minimum reliability of listed remailers.
A nil value means to list all remailers, regardless of reliability."
  :group 'mixmaster)

(defcustom mixmaster-max-latency nil
  "Maximum latency of listed remailers, in minutes.
A nil value means to list all remailers, regardless of latency."
  :group 'mixmaster)

;;}}}
;;{{{ Minor Mode setup

;;; Setting up the keymaps.
(defvar mixmaster-minor-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map [(control c) m a] 'mix-attach-file)
    (define-key map [(control c) m l] 'mix-set-latency)
    (define-key map [(control c) m c] 'mix-set-copies)
    (define-key map [(control c) m m] 'mixmaster-send)
    map)
  "Keyboard map for `mixmaster-minor-mode'.")

(defvar mixmaster-minor-mode-string " Mix"
  "*String to pu in the mode line when Mixmaster minor mode is active.")

;; Install our minor mode.
(or (assq 'mixmaster-minor-mode minor-mode-map-alist)
    (setq minor-mode-map-alist
          (cons (cons 'mixmaster-minor-mode mixmaster-minor-mode-map)
                minor-mode-map-alist)))

(or (assq 'mixmaster-minor-mode minor-mode-alist)
    (setq minor-mode-alist
          (cons '(mixmaster-minor-mode mixmaster-minor-mode-string)
                minor-mode-alist)))

(defvar mixmaster-minor-mode nil
  "If t, Mixmaster minor mode is active")

(make-variable-buffer-local 'mixmaster-minor-mode)

;; Keep track of our current mail sending function
(defvar mix-mail-send-and-exit 'mail-send-and-exit
  "This is the current mail sending function, normally bound to C-c C-c")

(make-variable-buffer-local mix-mail-send-and-exit)

;;;###autoload
(defun mixmaster-minor-mode (&optional arg)
  "\nMinor mode for sending mail and news through the Mixmaster remailer client.
\\<mixmaster-minor-mode-map>
\\[mixmaster-send]\t\tSend mail via Mixmaster
\\[mix-set-copies]\t\tSet the desired number of copies
\\[mix-attach-file]\t\tAttach a file via Mixmaster
\\[mix-set-latency]\t\tRequest the desired remailer latency\n"
  (interactive)
  (setq mixmaster-minor-mode
        (if (null arg) (not mixmaster-minor-mode)
          (> (prefix-numeric-value arg) 0))))

;;}}}
;;{{{ VM and GNUS interaction

(defun mixmaster-vm-setup (&optional arg)
  "Configure the VM mail editor to use Mixmaster."
  (interactive)
  (setq mix-mail-send-and-exit 'vm-mail-send-and-exit)
  (mixmaster-minor-mode t))

(defun mixmaster-gnus-setup (&optional arg)
  "Configure GNUS to use Mixmaster for posting anonymous usenet news."
  (interactive)
  (setq mix-mail-send-and-exit 'message-send-and-exit)
  (mixmaster-minor-mode t))

;;;###autoload
(defun mixmaster-send (&optional arg)
  "This is a wrapper around `mail-send-and-exit' which temporarily binds
`send-mail-function' to `mixmaster-send-it'."
  (interactive "P")
  (let ((send-mail-function 'mixmaster-send-it)
        (message-send-news-function 'mixmaster-send-it))
    (funcall mix-mail-send-and-exit arg)))

;;}}}
;;{{{ Calling Mixmaster

;;;###autoload
(defun mixmaster-send-it (&optional arg)
  "Send current mail message via Mixmaster."
  (interactive "P")
  (let ((mailbuf (current-buffer))
        (errbuf (get-buffer-create " mixmaster errors"))
        (tempbuf (find-file-noselect (make-temp-name (expand-file-name "~/tmp/emmix"))))
        (available-remailers (mix-full-remailer-list))
        mix-args mail-to mail-subject mail-body remailer-list)
    (loop for remailer = (completing-read
                          (format "Choose a remailer or press RET when finished (%s):"
                                  (cond ((listp remailer-list)
                                         (join-list remailer-list ","))
                                        (t
                                         "")))
                          available-remailers
                          'mix-remailer-filter)
      while (> (length remailer) 0)
      do (setq remailer-list (append remailer-list (list remailer))))
    (save-excursion
      (message "Preparing mail buffer for Mixmaster ...")
      (set-buffer mailbuf)

      (setq mail-subject (mail-fetch-field "Subject")
            mail-body (buffer-substring-no-properties
                       (mail-text)
                       (point-max)
                       mailbuf))

      ;; Save our message body.
      (set-buffer tempbuf)
      (insert mail-body)
      (save-buffer)
      ;; Set the file permissions so nobody can see what we're sending.
      (set-file-modes (buffer-file-name tempbuf) #o600)
      (set-buffer mailbuf)

      ;; Set up our mixmaster args
      (message "Calling Mixmaster ...")
      (setq mix-args
            (append (mix-make-mixmaster-args mailbuf)
                    (list (format "--subject=%s" mail-subject)
                          (format "--chain=%s" (join-list remailer-list ",")))))

      ;; Is it a mail message, or a news post?
      (cond ((mail-fetch-field "To")
             (setq mix-args (append mix-args
                                    (list "--mail"
                                          (format "--to=%s"
                                                  (mail-fetch-field "To"))))))
            ((mail-fetch-field "Newsgroups")
             (setq mix-args (append mix-args
                                    (list "--post"
                                          (format "--post-to=%s"
                                                  (mail-fetch-field "Newsgroups")))))))

      ;; Before calling mixmaster, we want to make sure our output
      ;; buffer is empty.  It's a risk to leave that information
      ;; sitting around for long periods of time.  And besides, if you
      ;; are remailing a lot, it fills up!
      (save-excursion
        (set-buffer errbuf)
        (erase-buffer))
      
      ;; Call the mixmaster process.  Need to use `apply' here because
      ;; call-process doesn't like taking a list as the final argument.
      (apply 'call-process mixmaster-binary
             (buffer-file-name tempbuf)
             errbuf nil
             mix-args)
      ;; Now that this is done, delete our message!
      (delete-file (buffer-file-name tempbuf))
      (kill-buffer tempbuf)
      (message "Mail sent!"))))

(defun mix-make-mixmaster-args (&optional buf)
  (let ((mix-args (list "-M" "--send" "--verbose")))
    (if (null buf)
        (setq buf (current-buffer)))
    (save-excursion
      (set-buffer buf)
      ;; Now let's build our command line.  These various headers are
      ;; still in the beta stages, but I think something like this
      ;; should work pretty well.
      (if (mail-fetch-field "X-Nym")
          (setq mix-args
                (append mix-args
                        (list (format "--nym=%s"
                                      (mail-fetch-field "X-Nym"))))))
      (if (mail-fetch-field "X-Attach")
          (let ((filename (expand-file-name (mail-fetch-field "X-Attach"))))
            (if (file-exists-p filename)
                (setq mix-args
                      (append mix-args
                              (list (format "--attachment=%s" filename))))
              (message (format "Mixmaster warning: %s not found!" filename)))))
      (if (mail-fetch-field "X-Latency")
          (setq mix-args
                (append mix-args
                        (list (format "--latency=%s"
                                      (mail-fetch-field "X-Latency"))))))
      (if (mail-fetch-field "X-Copies")
          (setq mix-args
                (append mix-args
                        (list (format "--copies=%s"
                                      (mail-fetch-field "X-Copies")))))))
    mix-args))

;;}}}
;;{{{ Header creation

(defun mix-attach-file (filename)
  "Use Mixmaster to attach a file to the current mail message."
  (interactive "fAttach File: ")
  (mix-set-header "X-Attach" (expand-file-name filename)))

(defun mix-set-latency (hours)
  "Request of Mixmaster that the message have a specified latency."
  (interactive "nRequested latency, in hours: ")
  (mix-set-header "X-Latency" (int-to-string hours)))

(defun mix-set-copies (copies)
  "Ask Mixmaster to send a specified number of hours.
This is only useful if you are using random remailers."
  (interactive "nNumber of copies to be sent: ")
  (mix-set-header "X-Copies" (int-to-string copies)))

(defun mix-set-header (header value)
  (save-excursion
    (if (mail-fetch-field header)
        (let (start end)
          (mail-position-on-field header)
          (setq end (point))
          (re-search-backward ": ")
          (setq start (+ (point) 2))
          (delete-region start end)))
    (mail-position-on-field header)
    (insert value)))

;;}}}
;;{{{ Info and Support

(defun mix-remailer-list (&optional type2-list)
  (unless type2-list
    (setq type2-list (concat mixmaster-directory "/type2.list")))
  (let ((buf (find-file-noselect type2-list))
        list)
    (save-excursion
      (set-buffer buf)
      (goto-char (point-min))
      (setq list (loop while (search-forward-regexp "^\\([a-z0-9]+\\) \\([^ ]+\\) .*$" nil t)
                   collect (cons (match-string 1) (match-string 2))))
      (kill-buffer buf))
    list))

(defun mix-full-remailer-list (&optional mlist-txt)
  "Parse an mlist2.txt remailer statistics file and return an alist with members
like this:

   (REMAILER (OPTIONS) RELIABILITY LATENCY)

Where REMAILER is a string with the name of the remailer, OPTIONS is a
list of symbols describing various settings for this particular
remailer, RELIABILITY is a floating point number representing the
reliability of the remailer (with 100.0 being the best), and LATENCY
is the number of minutes on average a message takes to make it through
this remailer."
  (if (null mlist-txt)
      (setq mlist-txt (expand-file-name "mlist2.txt" mixmaster-directory)))
  (if (not (file-readable-p mlist-txt))
      (error (format "Cannot read mlist2.txt from %s" mlist-txt)))
  (let ((mybuf (find-file-noselect mlist-txt))
        alist)
    (save-excursion
      (set-buffer mybuf)
      (goto-char (point-min))

      ;; Find the end of the mlist.txt header.
      (re-search-forward "^--------------------------------------------$" nil t)
      (forward-line 1)
      (beginning-of-line)

      ;; Collect statistics.
      (setq alist (loop while (re-search-forward
                               (concat
                                "^\\([a-z0-9]+\\) +"        ; remailer name
                                "[\\?0-9]+ +"               ; latent-hist
                                "\\([0-9]*:[0-9][0-9]\\) +" ; latency
                                "[+0-9\\?]+ +"              ; uptime-hist
                                "\\([0-9]+\\.[0-9]+\\)%  "  ; reliability
                                "\\([ a-zA-Z0-9]+\\)$")     ; options
                               nil t)
                    collect (list (match-string 1)
                                  (mix-option-list  (match-string 4))
                                  (string-to-number (match-string 3))
                                  (mix-calc-latency (match-string 2)))))
      (kill-buffer mybuf))
    alist))

(defun mix-calc-latency (latstr)
  (let (elements)
    (cond ((stringp latstr)
           (setq elements (split-string latstr ":"))
           (if (>= (length elements) 2)
               (+ (* (string-to-int (nth 0 elements)) 60)
                  (string-to-int (nth 1 elements)))
             0))
          (t
           0))))

(defun mix-option-list (option-string)
  "Parse an option string and return a capability list."
  (loop for char in (string-to-char-list option-string)
    with column = 0
    do (incf column)
    when (not (= char ?\ ))
    collect (cond ((= char ?D) "middle")
                  ((= char ?P) "post")
                  ((= char ?M) "mix")
                  ((= char ?R) "remix")
                  ((= char ?H) "hybrid")
                  ((= char ?G) "repgp")
                  ((= char ?O) "pgponly")
                  ((= char ?X) "ext")
                  ((= char ?A) "max")
                  ((= char ?T) "test")
                  ((= char ?L) "latent")
                  ((= char ?e) "ek")
                  ((= char ?E) "ekx")
                  ((= char ?U) "esub")
                  ((= char ?I) "inflt")
                  ((= char ?N) "rhop")
                  ((= column 15)
                   (format "klen%c" char))
                  ((= char ?2)
                   (if (= column 3)
                       "remix2"
                     "repgp2"))
                  (t
                   (char-to-string char)))))

(defun mix-remailer-filter (item)
  (not (or (if (and mixmaster-min-reliable
                    (< (caddr item) mixmaster-min-reliable))
               t)
           (if (and mixmaster-max-latency
                    (> (cadddr item) mixmaster-max-latency))
               t)
           nil)))

;;}}}
;;{{{ Utility functions

(defun join-list (jlist &optional sep)
  "Join a list of strings together into a single string, delimited by SEP."
  (mapconcat #'identity
             jlist (or sep " ")))

;;}}}
;;{{{ Provides

(provide 'mixmaster)

;;}}}

;; Local variables:
;; folded-file: t
;; folding-internal-margins: nil
;; time-stamp-start: "Last-Modified:[     ]+\\\\?[\"<]+"
;; time-stamp-end: "\\\\?[\">]"
;; time-stamp-line-limit: 10
;; time-stamp-format: "%4y-%02m-%02d %02H:%02M:%02S (%u)"
;; end:
